@@ -1,7 +1,6 @@
-#![doc(html_root_url = "https://docs.rs/singleton-trait/0.1.2")]
+#![doc(html_root_url = "https://docs.rs/singleton-trait/0.2.0")]
 #![no_std]
 
-
 use core::cell::{Cell, UnsafeCell};
 use core::marker::PhantomData;
 use core::mem::ManuallyDrop;
@@ -63,7 +62,7 @@ unsafe impl<'a, T: Singleton> Singleton for Erased<T> {}
 /*********************/
 
 /**
- * The Exists struct witnesses the logical ownership of a value of type T
+ * The Erased struct witnesses the logical ownership of a value of type T
  * while remaining zero-sized. This can be used for ghost proofs of soundness.
  *
  * Erased<T> should be thought of a zero-sized owner of T.
@@ -72,10 +71,12 @@ unsafe impl<'a, T: Singleton> Singleton for Erased<T> {}
  * of a valid T, which might not be true if they were called. On the other hand, since
  * it does not hold a T, it cannot drop T when it is itself dropped
  *
- * Secondly, keep in mind that while Exists<T> serves as evidence, it does not include
- * sufficient providence for either Stacked Borrows or LLVM, and so it is not sound
- * in general to recover a reference from an `Exists<&T>` and `*mut T` or `&UnsafeCell<T>`
+ * Secondly, keep in mind that while Erased<T> serves as evidence, it does not include
+ * sufficient provenance for Stacked Borrows or LLVM, and so it is not usually sound
+ * to recover a reference `&T` from an `Exists<&T>` and `*mut T` or `&UnsafeCell<T>`
  * even when T is Singleton, but this could be possible if T is zero-sized.
+ * Because of the missing provenance, creating a reference this way could invalidate
+ * the original reference on which this Exists instance was based.
  */
 #[derive(Clone, Copy)]
 pub struct Erased<T> {
@@ -106,12 +107,20 @@ impl<T> Erased<T> {
         }
     }
 
+    /**
+     * Turns a &Erased<T> into an Erased<&T>
+     */
+    #[inline(always)]
     pub fn borrow(&self) -> Erased<&T> {
         // Safety:
         // the identity function is pure
         unsafe { self.map_borrow(|r| r) }
     }
 
+    /**
+     * Turns a &mut Erased<T> into an Erased<&mut T>
+     */
+    #[inline(always)]
     pub fn borrow_mut(&mut self) -> Erased<&mut T> {
         // Safety:
         // the identity function is pure
@@ -119,7 +128,7 @@ impl<T> Erased<T> {
     }
 
     /**
-     * Maps a function on the inside of the Erased body (existential induction).
+     * Maps a function on the inside of the Erased field.
      *
      * # Safety
      *
@@ -135,7 +144,7 @@ impl<T> Erased<T> {
     }
 
     /**
-     * Maps a function on the borrow of the Erased field (existential induction).
+     * Maps a function on the borrow of the Erased field.
      *
      * # Safety
      *
@@ -170,12 +179,35 @@ impl<T> Erased<T> {
         // By the contract for the passed function, this is equivalent to calling it on the mutable borrow of T
         Erased::<R>::new_unchecked()
     }
+
+    /**
+     * Fallback function which converts this Erased value into an Exists
+     * implementer.
+     *
+     * This exists because we cannot allow general trait implementations
+     * due to coherence rules, but we can't be sufficiently flexible in
+     * trait implementations due to a lack of subtyping constraints or
+     * trait covariance, which would mean references would be too inflexible
+     */
+    #[inline(always)]
+    pub fn exists(self) -> impl Exists<T> {
+        struct Internal<T> {
+            er: Erased<T>,
+        }
+        impl<T> Exists<T> for Internal<T> {
+            fn erase(self) -> Erased<T> {
+                self.er
+            }
+        }
+        Internal { er: self }
+    }
 }
 impl<T> Erased<Erased<T>> {
     /**
      * An erased erased value can be flattened into a single erasure,
      * since Erased<T> is notionally equivalent to T
      */
+    #[inline(always)]
     pub fn flatten(self) -> Erased<T> {
         // SAFETY:
         //
@@ -184,6 +216,7 @@ impl<T> Erased<Erased<T>> {
     }
 }
 impl<'a, T> Erased<&'a mut T> {
+    #[inline(always)]
     pub fn reborrow<'b>(&'b mut self) -> Erased<&'b mut T> {
         // SAFETY
         //
@@ -199,7 +232,92 @@ impl<T> From<T> for Erased<T> {
 /**
  * The Exists trait is intended to be used with `impl`, to denote
  * an argument where the existence of a value is sufficient as an argument
+ *
+ *
  */
-pub trait Exists<T>: Into<Erased<T>> {}
-impl<T: Into<Erased<T>>> Exists<T> for T {}
+pub trait Exists<T: Sized> {
+    fn erase(self) -> Erased<T>;
+}
+impl<T> Exists<T> for T {
+    #[inline(always)]
+    fn erase(self) -> Erased<T> {
+        self.into()
+    }
+}
 
+impl<'a, 'b: 'a, T> Exists<&'a T> for Erased<&'b T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map(|r: &'a T| &*r) }
+    }
+}
+impl<'a, 'b: 'a, T> Exists<&'a T> for Erased<&'b mut T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map(|r: &'a mut T| &*r) }
+    }
+}
+impl<'a, 'b: 'a, T> Exists<&'a mut T> for Erased<&'b mut T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a mut T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map(|r: &'a mut T| &mut *r) }
+    }
+}
+impl<'a, 'b, T> Exists<&'a T> for &'a Erased<&'b T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map(|r: &'a T| &*r) }
+    }
+}
+impl<'a, 'b: 'a, T> Exists<&'a T> for &'a Erased<&'b mut T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map_borrow(|r: &'a &'b mut T| &**r) }
+    }
+}
+impl<'a, 'b: 'a, T> Exists<&'a mut T> for &'a mut Erased<&'b mut T> {
+    #[inline(always)]
+    fn erase(self) -> Erased<&'a mut T> {
+        // SAFETY: Deref on reference is pure
+        unsafe { self.map_borrow_mut(|r: &'a mut &'b mut T| &mut **r) }
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use crate::*;
+
+    // this test should fail to compile if we're missing impls
+    // that can coerce between different lifetimes
+    #[test]
+    fn test_impls() {
+        /* Invariant type parameter to strictly test subtyping */
+        struct NonCopy;
+        impl Drop for NonCopy {
+            fn drop(&mut self) {}
+        }
+        struct Guard<'a>(PhantomData<fn(&'a ()) -> &'a ()>, &'a NonCopy);
+
+        fn takes_ref<'a>(_: &Guard<'a>, _: impl Exists<&'a ()>) {}
+        fn takes_mut<'a>(_: &Guard<'a>, _: impl Exists<&'a mut ()>) {}
+
+        let nc = NonCopy;
+        let guard = Guard(PhantomData, &nc);
+
+        let er = Erased::new(&());
+        takes_ref(&guard, er);
+
+        let mut x = ();
+        let er = Erased::new(&mut x);
+        takes_ref(&guard, er);
+
+        let mut x = ();
+        let er = Erased::new(&mut x);
+        takes_mut(&guard, er);
+    }
+}
